import tomllib
from pathlib import Path

import tomli_w

CONFIG_DIR = Path.home() / ".mtg_print"
CONFIG_FILE = CONFIG_DIR / "config.toml"


def load_preferences() -> dict[str, str]:
    if not CONFIG_FILE.exists():
        return {}

    with CONFIG_FILE.open("rb") as f:
        config = tomllib.load(f)

    return config.get("preferences", {})


def save_preference(card_name: str, set_code: str) -> None:
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    preferences = load_preferences()
    preferences[card_name] = set_code.lower()

    config = {"preferences": preferences}

    with CONFIG_FILE.open("wb") as f:
        tomli_w.dump(config, f)


def get_preference(card_name: str) -> str | None:
    preferences = load_preferences()
    return preferences.get(card_name)


def clear_preferences() -> int:
    if not CONFIG_FILE.exists():
        return 0

    preferences = load_preferences()
    count = len(preferences)

    if CONFIG_FILE.exists():
        CONFIG_FILE.unlink()

    return count



from dataclasses import dataclass, field
from datetime import date


@dataclass
class CardFace:
    name: str
    image_uri_png: str
    image_uri_small: str | None = None


@dataclass
class CardPrinting:
    name: str
    set_code: str
    set_name: str
    collector_number: str
    release_date: date
    scryfall_uri: str
    layout: str
    faces: list[CardFace] = field(default_factory=list)

    @property
    def is_double_faced(self) -> bool:
        return self.layout in ("transform", "modal_dfc", "reversible_card")

    @property
    def front_image(self) -> str:
        return self.faces[0].image_uri_png

    @property
    def back_image(self) -> str | None:
        return self.faces[1].image_uri_png if len(self.faces) > 1 else None


@dataclass
class DeckEntry:
    count: int
    name: str
    set_override: str | None = None


@dataclass
class Decklist:
    entries: list[DeckEntry] = field(default_factory=list)

    @property
    def total_cards(self) -> int:
        return sum(e.count for e in self.entries)



from datetime import date

import httpx
import pytest
from pytest_httpx import HTTPXMock

from mtg_print.scryfall import CardNotFoundError, ScryfallClient

ELVISH_RECLAIMER_RESPONSE = {
    "name": "Elvish Reclaimer",
    "set": "mh1",
    "set_name": "Modern Horizons",
    "collector_number": "158",
    "released_at": "2019-06-14",
    "scryfall_uri": "https://scryfall.com/card/mh1/158",
    "layout": "normal",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/mh1/158.png",
        "small": "https://cards.scryfall.io/small/mh1/158.jpg",
    },
}

DELVER_DFC_RESPONSE = {
    "name": "Delver of Secrets // Insectile Aberration",
    "set": "isd",
    "set_name": "Innistrad",
    "collector_number": "51",
    "released_at": "2011-09-30",
    "scryfall_uri": "https://scryfall.com/card/isd/51",
    "layout": "transform",
    "card_faces": [
        {
            "name": "Delver of Secrets",
            "image_uris": {
                "png": "https://cards.scryfall.io/png/isd/51a.png",
                "small": "https://cards.scryfall.io/small/isd/51a.jpg",
            },
        },
        {
            "name": "Insectile Aberration",
            "image_uris": {
                "png": "https://cards.scryfall.io/png/isd/51b.png",
                "small": "https://cards.scryfall.io/small/isd/51b.jpg",
            },
        },
    ],
}

SEARCH_RESPONSE = {
    "data": [
        ELVISH_RECLAIMER_RESPONSE,
        {
            **ELVISH_RECLAIMER_RESPONSE,
            "set": "mh2",
            "set_name": "Modern Horizons 2",
            "collector_number": "166",
            "released_at": "2021-06-18",
        },
    ]
}


class TestScryfallClientParsePrinting:
    def test_parses_normal_card(self) -> None:
        client = ScryfallClient()
        printing = client._parse_printing(ELVISH_RECLAIMER_RESPONSE)

        assert printing.name == "Elvish Reclaimer"
        assert printing.set_code == "mh1"
        assert printing.set_name == "Modern Horizons"
        assert printing.collector_number == "158"
        assert printing.release_date == date(2019, 6, 14)
        assert printing.layout == "normal"
        assert len(printing.faces) == 1
        assert printing.faces[0].name == "Elvish Reclaimer"
        assert printing.faces[0].image_uri_png == "https://cards.scryfall.io/png/mh1/158.png"

    def test_parses_double_faced_card(self) -> None:
        client = ScryfallClient()
        printing = client._parse_printing(DELVER_DFC_RESPONSE)

        assert printing.name == "Delver of Secrets // Insectile Aberration"
        assert printing.layout == "transform"
        assert printing.is_double_faced is True
        assert len(printing.faces) == 2
        assert printing.faces[0].name == "Delver of Secrets"
        assert printing.faces[1].name == "Insectile Aberration"
        assert printing.front_image == "https://cards.scryfall.io/png/isd/51a.png"
        assert printing.back_image == "https://cards.scryfall.io/png/isd/51b.png"

    def test_normal_card_not_double_faced(self) -> None:
        client = ScryfallClient()
        printing = client._parse_printing(ELVISH_RECLAIMER_RESPONSE)

        assert printing.is_double_faced is False
        assert printing.back_image is None


class TestScryfallClientGetCardByName:
    def test_empty_search_results_raises_error(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(json={"data": []})

        client = ScryfallClient()
        with pytest.raises(CardNotFoundError) as exc_info:
            client.get_card_by_name("Obscure Card")

        assert exc_info.value.card_name == "Obscure Card"

    def test_with_set_code_uses_named_endpoint(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Elvish+Reclaimer&set=mh1",
            json=ELVISH_RECLAIMER_RESPONSE,
        )

        client = ScryfallClient()
        printing = client.get_card_by_name("Elvish Reclaimer", set_code="MH1")

        assert printing.name == "Elvish Reclaimer"
        assert printing.set_code == "mh1"

    def test_without_set_code_uses_search(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/search?q=%21%22Elvish+Reclaimer%22+include%3Aextras&unique=prints&order=released&dir=asc",
            json=SEARCH_RESPONSE,
        )

        client = ScryfallClient()
        printing = client.get_card_by_name("Elvish Reclaimer")

        assert printing.name == "Elvish Reclaimer"
        assert printing.set_code == "mh1"  # First (oldest) printing

    def test_card_not_found_raises_error(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Not+A+Real+Card&set=xxx",
            status_code=404,
        )

        client = ScryfallClient()
        with pytest.raises(CardNotFoundError) as exc_info:
            client.get_card_by_name("Not A Real Card", set_code="xxx")

        assert exc_info.value.card_name == "Not A Real Card"
        assert "Card not found" in str(exc_info.value)


class TestScryfallClientSearchPrintings:
    def test_returns_all_printings_oldest_first(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/search?q=%21%22Elvish+Reclaimer%22+include%3Aextras&unique=prints&order=released&dir=asc",
            json=SEARCH_RESPONSE,
        )

        client = ScryfallClient()
        printings = client.search_printings("Elvish Reclaimer")

        assert len(printings) == 2
        assert printings[0].set_code == "mh1"  # Older
        assert printings[1].set_code == "mh2"  # Newer

    def test_card_not_found_raises_error(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/search?q=%21%22Fake+Card%22+include%3Aextras&unique=prints&order=released&dir=asc",
            status_code=404,
        )

        client = ScryfallClient()
        with pytest.raises(CardNotFoundError):
            client.search_printings("Fake Card")

    def test_filters_out_art_series_without_images(self, httpx_mock: HTTPXMock) -> None:
        response_with_art_series = {
            "data": [
                {
                    "name": "Wrenn and Six // Wrenn and Six",
                    "set": "amh1",
                    "set_name": "Modern Horizons Art Series",
                    "collector_number": "6",
                    "released_at": "2019-06-05",
                    "scryfall_uri": "https://scryfall.com/card/amh1/6",
                    "layout": "art_series",
                    # No image_uris - art series cards don't have them
                },
                {
                    "name": "Wrenn and Six",
                    "set": "mh1",
                    "set_name": "Modern Horizons",
                    "collector_number": "217",
                    "released_at": "2019-06-14",
                    "scryfall_uri": "https://scryfall.com/card/mh1/217",
                    "layout": "normal",
                    "image_uris": {
                        "png": "https://cards.scryfall.io/png/mh1/217.png",
                        "small": "https://cards.scryfall.io/small/mh1/217.jpg",
                    },
                },
            ]
        }
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/search?q=%21%22Wrenn+and+Six%22+include%3Aextras&unique=prints&order=released&dir=asc",
            json=response_with_art_series,
        )

        client = ScryfallClient()
        printings = client.search_printings("Wrenn and Six")

        assert len(printings) == 1
        assert printings[0].set_code == "mh1"
        assert printings[0].layout == "normal"


class TestScryfallClientDownloadImage:
    def test_downloads_and_saves_image(self, httpx_mock: HTTPXMock, tmp_path) -> None:
        fake_png_data = b"\x89PNG\r\n\x1a\n" + b"fake image data"
        httpx_mock.add_response(
            url="https://cards.scryfall.io/png/mh1/158.png",
            content=fake_png_data,
        )

        client = ScryfallClient()
        dest = tmp_path / "mh1" / "158.png"
        result = client.download_image("https://cards.scryfall.io/png/mh1/158.png", dest)

        assert result == dest
        assert dest.exists()
        assert dest.read_bytes() == fake_png_data

    def test_creates_parent_directories(self, httpx_mock: HTTPXMock, tmp_path) -> None:
        httpx_mock.add_response(
            url="https://cards.scryfall.io/png/mh1/158.png",
            content=b"image",
        )

        client = ScryfallClient()
        dest = tmp_path / "deep" / "nested" / "path" / "158.png"
        client.download_image("https://cards.scryfall.io/png/mh1/158.png", dest)

        assert dest.exists()


class TestScryfallClientCustomHttpClient:
    def test_accepts_custom_http_client(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(json=ELVISH_RECLAIMER_RESPONSE)

        custom_client = httpx.Client(timeout=60.0)
        client = ScryfallClient(http_client=custom_client)
        printing = client.get_card_by_name("Elvish Reclaimer", set_code="mh1")

        assert printing.name == "Elvish Reclaimer"


AVENGER_RESPONSE = {
    "name": "Avenger of Zendikar",
    "set": "c18",
    "set_name": "Commander 2018",
    "collector_number": "129",
    "released_at": "2018-08-10",
    "scryfall_uri": "https://scryfall.com/card/c18/129",
    "layout": "normal",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/c18/129.png",
        "small": "https://cards.scryfall.io/small/c18/129.jpg",
    },
    "all_parts": [
        {
            "object": "related_card",
            "id": "abc123",
            "component": "combo_piece",
            "name": "Avenger of Zendikar",
            "uri": "https://api.scryfall.com/cards/abc123",
        },
        {
            "object": "related_card",
            "id": "plant123",
            "component": "token",
            "name": "Plant",
            "uri": "https://api.scryfall.com/cards/plant123",
        },
    ],
}

PLANT_TOKEN_RESPONSE = {
    "name": "Plant",
    "set": "tc18",
    "set_name": "Commander 2018 Tokens",
    "collector_number": "1",
    "released_at": "2018-08-10",
    "scryfall_uri": "https://scryfall.com/card/tc18/1",
    "layout": "token",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/tc18/1.png",
        "small": "https://cards.scryfall.io/small/tc18/1.jpg",
    },
}

TEFERI_RESPONSE = {
    "name": "Teferi, Hero of Dominaria",
    "set": "dom",
    "set_name": "Dominaria",
    "collector_number": "207",
    "released_at": "2018-04-27",
    "scryfall_uri": "https://scryfall.com/card/dom/207",
    "layout": "normal",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/dom/207.png",
        "small": "https://cards.scryfall.io/small/dom/207.jpg",
    },
    "all_parts": [
        {
            "object": "related_card",
            "id": "emblem123",
            "component": "combo_piece",
            "name": "Teferi, Hero of Dominaria Emblem",
            "uri": "https://api.scryfall.com/cards/emblem123",
        },
        {
            "object": "related_card",
            "id": "teferi123",
            "component": "combo_piece",
            "name": "Teferi, Hero of Dominaria",
            "uri": "https://api.scryfall.com/cards/teferi123",
        },
    ],
}

TEFERI_EMBLEM_RESPONSE = {
    "name": "Teferi, Hero of Dominaria Emblem",
    "set": "tdom",
    "set_name": "Dominaria Tokens",
    "collector_number": "1",
    "released_at": "2018-04-27",
    "scryfall_uri": "https://scryfall.com/card/tdom/1",
    "layout": "emblem",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/tdom/1.png",
        "small": "https://cards.scryfall.io/small/tdom/1.jpg",
    },
}

GISELA_RESPONSE = {
    "name": "Gisela, the Broken Blade",
    "set": "emn",
    "set_name": "Eldritch Moon",
    "collector_number": "28",
    "released_at": "2016-07-22",
    "scryfall_uri": "https://scryfall.com/card/emn/28",
    "layout": "meld",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/emn/28.png",
        "small": "https://cards.scryfall.io/small/emn/28.jpg",
    },
    "all_parts": [
        {
            "object": "related_card",
            "id": "checklist123",
            "component": "token",
            "name": "Eldritch Moon Checklist",
            "uri": "https://api.scryfall.com/cards/checklist123",
        },
        {
            "object": "related_card",
            "id": "brisela123",
            "component": "meld_result",
            "name": "Brisela, Voice of Nightmares",
            "uri": "https://api.scryfall.com/cards/brisela123",
        },
        {
            "object": "related_card",
            "id": "gisela123",
            "component": "meld_part",
            "name": "Gisela, the Broken Blade",
            "uri": "https://api.scryfall.com/cards/gisela123",
        },
    ],
}

BRISELA_RESPONSE = {
    "name": "Brisela, Voice of Nightmares",
    "set": "emn",
    "set_name": "Eldritch Moon",
    "collector_number": "15b",
    "released_at": "2016-07-22",
    "scryfall_uri": "https://scryfall.com/card/emn/15b",
    "layout": "meld",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/emn/15b.png",
        "small": "https://cards.scryfall.io/small/emn/15b.jpg",
    },
}

NO_PARTS_RESPONSE = {
    **ELVISH_RECLAIMER_RESPONSE,
}

MINSC_RESPONSE = {
    "name": "Minsc & Boo, Timeless Heroes",
    "set": "clb",
    "set_name": "Commander Legends: Battle for Baldur's Gate",
    "collector_number": "285",
    "released_at": "2022-06-10",
    "scryfall_uri": "https://scryfall.com/card/clb/285",
    "layout": "normal",
    "image_uris": {
        "png": "https://cards.scryfall.io/png/clb/285.png",
        "small": "https://cards.scryfall.io/small/clb/285.jpg",
    },
    "all_parts": [
        {
            "object": "related_card",
            "id": "minsc123",
            "component": "combo_piece",
            "name": "Minsc & Boo, Timeless Heroes",
            "uri": "https://api.scryfall.com/cards/minsc123",
        },
        {
            "object": "related_card",
            "id": "boo123",
            "component": "token",
            "name": "Boo",
            "uri": "https://api.scryfall.com/cards/boo123",
        },
        {
            "object": "related_card",
            "id": "aminsc123",
            "component": "combo_piece",
            "name": "A-Minsc & Boo, Timeless Heroes",
            "uri": "https://api.scryfall.com/cards/aminsc123",
        },
    ],
}

BOO_TOKEN_RESPONSE = {
    "name": "Boo",
    "set": "tclb",
    "set_name": "Commander Legends: Battle for Baldur's Gate Tokens",
    "collector_number": "1",
    "released_at": "2022-06-10",
    "scryfall_uri": "https://scryfall.com/card/tclb/1",
    "layout": "token",
    "digital": False,
    "image_uris": {
        "png": "https://cards.scryfall.io/png/tclb/1.png",
        "small": "https://cards.scryfall.io/small/tclb/1.jpg",
    },
}

ALCHEMY_MINSC_RESPONSE = {
    "name": "A-Minsc & Boo, Timeless Heroes",
    "set": "hbg",
    "set_name": "Alchemy Horizons: Baldur's Gate",
    "collector_number": "285",
    "released_at": "2022-07-07",
    "scryfall_uri": "https://scryfall.com/card/hbg/285",
    "layout": "normal",
    "digital": True,
    "image_uris": {
        "png": "https://cards.scryfall.io/png/hbg/285.png",
        "small": "https://cards.scryfall.io/small/hbg/285.jpg",
    },
}


class TestScryfallClientGetRelatedParts:
    def test_returns_tokens_for_card(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Avenger+of+Zendikar",
            json=AVENGER_RESPONSE,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/plant123",
            json=PLANT_TOKEN_RESPONSE,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Avenger of Zendikar")

        assert len(parts) == 1
        assert parts[0].name == "Plant"
        assert parts[0].layout == "token"

    def test_returns_emblems_for_planeswalker(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Teferi%2C+Hero+of+Dominaria",
            json=TEFERI_RESPONSE,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/emblem123",
            json=TEFERI_EMBLEM_RESPONSE,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Teferi, Hero of Dominaria")

        assert len(parts) == 1
        assert parts[0].name == "Teferi, Hero of Dominaria Emblem"
        assert parts[0].layout == "emblem"

    def test_excludes_self_reference(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Teferi%2C+Hero+of+Dominaria",
            json=TEFERI_RESPONSE,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/emblem123",
            json=TEFERI_EMBLEM_RESPONSE,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Teferi, Hero of Dominaria")

        part_names = [p.name for p in parts]
        assert "Teferi, Hero of Dominaria" not in part_names

    def test_excludes_checklist_cards(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Gisela%2C+the+Broken+Blade",
            json=GISELA_RESPONSE,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/brisela123",
            json=BRISELA_RESPONSE,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Gisela, the Broken Blade")

        part_names = [p.name for p in parts]
        assert "Eldritch Moon Checklist" not in part_names
        assert "Brisela, Voice of Nightmares" in part_names

    def test_returns_empty_for_card_without_parts(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Elvish+Reclaimer",
            json=NO_PARTS_RESPONSE,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Elvish Reclaimer")

        assert parts == []

    def test_with_set_code(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Avenger+of+Zendikar&set=c18",
            json=AVENGER_RESPONSE,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/plant123",
            json=PLANT_TOKEN_RESPONSE,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Avenger of Zendikar", set_code="C18")

        assert len(parts) == 1
        assert parts[0].name == "Plant"

    def test_excludes_digital_only_cards(self, httpx_mock: HTTPXMock) -> None:
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Minsc+%26+Boo%2C+Timeless+Heroes",
            json=MINSC_RESPONSE,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/boo123",
            json=BOO_TOKEN_RESPONSE,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/aminsc123",
            json=ALCHEMY_MINSC_RESPONSE,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Minsc & Boo, Timeless Heroes")

        part_names = [p.name for p in parts]
        assert "Boo" in part_names
        assert "A-Minsc & Boo, Timeless Heroes" not in part_names
        assert len(parts) == 1

    def test_excludes_normal_cards_from_related_parts(self, httpx_mock: HTTPXMock) -> None:
        elk_response = {
            "name": "Elk",
            "set": "tecl",
            "set_name": "Throne of Eldraine Tokens",
            "collector_number": "1",
            "released_at": "2019-10-04",
            "scryfall_uri": "https://scryfall.com/card/tecl/1",
            "layout": "token",
            "image_uris": {
                "png": "https://cards.scryfall.io/png/tecl/1.png",
                "small": "https://cards.scryfall.io/small/tecl/1.jpg",
            },
            "all_parts": [
                {
                    "object": "related_card",
                    "id": "elk123",
                    "component": "token",
                    "name": "Elk",
                    "uri": "https://api.scryfall.com/cards/elk123",
                },
                {
                    "object": "related_card",
                    "id": "oko123",
                    "component": "combo_piece",
                    "name": "Oko, Thief of Crowns",
                    "uri": "https://api.scryfall.com/cards/oko123",
                },
            ],
        }
        oko_response = {
            "name": "Oko, Thief of Crowns",
            "set": "eld",
            "set_name": "Throne of Eldraine",
            "collector_number": "197",
            "released_at": "2019-10-04",
            "scryfall_uri": "https://scryfall.com/card/eld/197",
            "layout": "normal",
            "image_uris": {
                "png": "https://cards.scryfall.io/png/eld/197.png",
                "small": "https://cards.scryfall.io/small/eld/197.jpg",
            },
        }
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/named?exact=Elk",
            json=elk_response,
        )
        httpx_mock.add_response(
            url="https://api.scryfall.com/cards/oko123",
            json=oko_response,
        )

        client = ScryfallClient()
        parts = client.get_related_parts("Elk")

        part_names = [p.name for p in parts]
        assert "Oko, Thief of Crowns" not in part_names
        assert len(parts) == 0



import time
from datetime import date
from unittest.mock import MagicMock

import pytest

from mtg_print.cache import ImageCache
from mtg_print.models import CardFace, CardPrinting


def make_printing(
    set_code: str = "mh1",
    collector_number: str = "158",
    layout: str = "normal",
    faces: list[CardFace] | None = None,
) -> CardPrinting:
    if faces is None:
        faces = [
            CardFace(
                name="Test Card",
                image_uri_png=f"https://cards.scryfall.io/png/{set_code}/{collector_number}.png",
                image_uri_small=f"https://cards.scryfall.io/small/{set_code}/{collector_number}.jpg",
            )
        ]
    return CardPrinting(
        name="Test Card",
        set_code=set_code,
        set_name="Test Set",
        collector_number=collector_number,
        release_date=date(2020, 1, 1),
        scryfall_uri="https://scryfall.com/card/test/1",
        layout=layout,
        faces=faces,
    )


class TestImageCacheGetPath:
    def test_returns_correct_path_structure(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing(set_code="mh1", collector_number="158")

        path = cache.get_path(printing)

        assert path == tmp_path / "mh1" / "158.png"

    def test_back_face_has_suffix(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing(set_code="isd", collector_number="51")

        front_path = cache.get_path(printing, face_index=0)
        back_path = cache.get_path(printing, face_index=1)

        assert front_path == tmp_path / "isd" / "51.png"
        assert back_path == tmp_path / "isd" / "51_back.png"


class TestImageCacheHas:
    def test_returns_false_when_not_cached(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing()

        assert cache.has(printing) is False

    def test_returns_true_when_cached(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing(set_code="mh1", collector_number="158")

        (tmp_path / "mh1").mkdir()
        (tmp_path / "mh1" / "158.png").write_bytes(b"image data")

        assert cache.has(printing) is True

    def test_checks_correct_face(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing(set_code="isd", collector_number="51")

        (tmp_path / "isd").mkdir()
        (tmp_path / "isd" / "51.png").write_bytes(b"front")

        assert cache.has(printing, face_index=0) is True
        assert cache.has(printing, face_index=1) is False


class TestImageCacheGetOrDownload:
    def test_returns_cached_file_without_download(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing(set_code="mh1", collector_number="158")

        (tmp_path / "mh1").mkdir()
        cached_file = tmp_path / "mh1" / "158.png"
        cached_file.write_bytes(b"cached image")

        mock_client = MagicMock()
        result = cache.get_or_download(printing, mock_client)

        assert result == cached_file
        mock_client.download_image.assert_not_called()

    def test_downloads_when_not_cached(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing(set_code="mh1", collector_number="158")

        mock_client = MagicMock()
        expected_path = tmp_path / "mh1" / "158.png"
        mock_client.download_image.return_value = expected_path

        result = cache.get_or_download(printing, mock_client)

        mock_client.download_image.assert_called_once_with(
            "https://cards.scryfall.io/png/mh1/158.png",
            expected_path,
        )
        assert result == expected_path

    def test_touches_file_on_cache_hit(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing(set_code="mh1", collector_number="158")

        (tmp_path / "mh1").mkdir()
        cached_file = tmp_path / "mh1" / "158.png"
        cached_file.write_bytes(b"cached image")

        old_mtime = cached_file.stat().st_mtime
        time.sleep(0.01)

        mock_client = MagicMock()
        cache.get_or_download(printing, mock_client)

        new_mtime = cached_file.stat().st_mtime
        assert new_mtime > old_mtime

    def test_raises_for_invalid_face_index(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        printing = make_printing()  # Single face

        mock_client = MagicMock()
        with pytest.raises(ValueError, match="No face at index"):
            cache.get_or_download(printing, mock_client, face_index=5)


class TestImageCacheEviction:
    def test_evicts_oldest_files_when_over_limit(self, tmp_path) -> None:
        (tmp_path / "set1").mkdir()
        (tmp_path / "set2").mkdir()

        old_file = tmp_path / "set1" / "old.png"
        new_file = tmp_path / "set2" / "new.png"

        old_file.write_bytes(b"x" * 1000)
        time.sleep(0.01)
        new_file.write_bytes(b"x" * 1000)

        ImageCache(cache_dir=tmp_path, max_size_gb=0.000001)

        assert not old_file.exists()
        assert new_file.exists()

    def test_no_eviction_when_under_limit(self, tmp_path) -> None:
        (tmp_path / "set1").mkdir()
        file1 = tmp_path / "set1" / "card.png"
        file1.write_bytes(b"small file")

        ImageCache(cache_dir=tmp_path, max_size_gb=1.0)

        assert file1.exists()


class TestImageCacheClear:
    def test_removes_all_cached_files(self, tmp_path) -> None:
        (tmp_path / "set1").mkdir()
        (tmp_path / "set2").mkdir()
        (tmp_path / "set1" / "card1.png").write_bytes(b"image1")
        (tmp_path / "set2" / "card2.png").write_bytes(b"image2")

        cache = ImageCache(cache_dir=tmp_path)
        count = cache.clear()

        assert count == 2
        assert not (tmp_path / "set1" / "card1.png").exists()
        assert not (tmp_path / "set2" / "card2.png").exists()

    def test_returns_zero_for_empty_cache(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        count = cache.clear()

        assert count == 0


class TestImageCacheStats:
    def test_returns_file_count_and_size(self, tmp_path) -> None:
        (tmp_path / "set1").mkdir()
        (tmp_path / "set1" / "card1.png").write_bytes(b"x" * 100)
        (tmp_path / "set1" / "card2.png").write_bytes(b"x" * 200)

        cache = ImageCache(cache_dir=tmp_path)
        stats = cache.stats()

        assert stats["file_count"] == 2
        assert stats["total_size_bytes"] == 300

    def test_empty_cache_stats(self, tmp_path) -> None:
        cache = ImageCache(cache_dir=tmp_path)
        stats = cache.stats()

        assert stats["file_count"] == 0
        assert stats["total_size_bytes"] == 0



"""End-to-end tests using real decklists against Scryfall API.

These tests hit the real Scryfall API and are marked as slow.
Run with: pytest tests/test_e2e.py -v -s -m slow
Skip with: pytest tests/ -v (default)
"""

import time
from pathlib import Path

import pytest

from mtg_print.decklist import parse_decklist
from mtg_print.scryfall import ScryfallClient

FIXTURES_DIR = Path(__file__).parent / "fixtures"
DECKLIST_FIXTURES = ["modern.txt", "legacy.txt", "commander.txt"]


@pytest.fixture
def client() -> ScryfallClient:
    return ScryfallClient()


@pytest.mark.slow
@pytest.mark.parametrize("fixture", DECKLIST_FIXTURES)
class TestDecklistE2E:
    def test_parses_and_fetches_all_cards(self, client: ScryfallClient, fixture: str) -> None:
        decklist = parse_decklist(FIXTURES_DIR / fixture)
        format_name = fixture.replace(".txt", "").title()

        start = time.perf_counter()
        for entry in decklist.entries:
            printing = client.get_card_by_name(entry.name)
            # DFCs: entry.name (front face) is contained in printing.name ("Front // Back")
            assert entry.name in printing.name
            assert len(printing.faces) >= 1
        elapsed = time.perf_counter() - start

        print(f"\n{format_name}: fetched {len(decklist.entries)} cards in {elapsed:.2f}s")

    def test_collects_related_parts(self, client: ScryfallClient, fixture: str) -> None:
        decklist = parse_decklist(FIXTURES_DIR / fixture)
        format_name = fixture.replace(".txt", "").title()
        all_parts: dict[str, str] = {}

        start = time.perf_counter()
        for entry in decklist.entries:
            for part in client.get_related_parts(entry.name):
                all_parts[part.name] = part.layout
        elapsed = time.perf_counter() - start

        print(f"\n{format_name}: found {len(all_parts)} related parts in {elapsed:.2f}s")
        for name, layout in sorted(all_parts.items()):
            print(f"  - {name} ({layout})")



import re
from pathlib import Path

from mtg_print.models import DeckEntry, Decklist

PATTERNS = [
    re.compile(
        r"^(\d+)x?\s+(.+?)\s*\((\w+)\)\s*\d*\s*(?:\*F\*)?$"
    ),  # Moxfield/Arena/Manabox: 4 Card (SET) 123 *F*
    re.compile(r"^(\d+)x?\s+(.+?)\s*<(\w+)>$"),  # MTGGoldfish: 4 Card <set>
    re.compile(r"^(\d+)x?\s+(.+?)$"),  # Simple: 4 Card or 4x Card
]

IGNORE_PATTERNS = [
    re.compile(r"^(Main|Main Deck|Sideboard|Companion|Commander)", re.IGNORECASE),
    re.compile(r"^(Creatures?|Spells?|Lands?|Planeswalkers?|Artifacts?|Enchantments?)\s*\(\d+\)"),
    re.compile(r"^(Built with|Exported from|Shared via|//)"),
    re.compile(r"^\s*$"),
]

BASIC_LANDS = {
    "Forest",
    "Island",
    "Plains",
    "Mountain",
    "Swamp",
    "Wastes",
    "Snow-Covered Forest",
    "Snow-Covered Island",
    "Snow-Covered Plains",
    "Snow-Covered Mountain",
    "Snow-Covered Swamp",
    "Snow-Covered Wastes",
}


def is_basic_land(name: str) -> bool:
    return name in BASIC_LANDS


def should_ignore_line(line: str) -> bool:
    return any(p.match(line) for p in IGNORE_PATTERNS)


def parse_line(line: str) -> DeckEntry | None:
    line = line.strip()
    if should_ignore_line(line):
        return None

    for pattern in PATTERNS:
        if match := pattern.match(line):
            groups = match.groups()
            count = int(groups[0])
            name = groups[1].strip()
            set_code = groups[2] if len(groups) > 2 else None
            return DeckEntry(count=count, name=name, set_override=set_code)

    return None


def parse_decklist(path: Path, skip_basics: bool = True) -> Decklist:
    with open(path) as f:
        lines = f.readlines()
    return parse_decklist_string("\n".join(lines), skip_basics=skip_basics)


def parse_decklist_string(content: str, skip_basics: bool = True) -> Decklist:
    entries: list[DeckEntry] = []
    for line in content.splitlines():
        if entry := parse_line(line):
            if skip_basics and is_basic_land(entry.name):
                continue
            entries.append(entry)
    return Decklist(entries=entries)



from pathlib import Path
from typing import Annotated

import typer

from mtg_print.__version__ import __version__
from mtg_print.cache import ImageCache
from mtg_print.decklist import parse_decklist
from mtg_print.models import CardPrinting
from mtg_print.preferences import (
    clear_preferences,
    get_preference,
    load_preferences,
    save_preference,
)
from mtg_print.scryfall import CardNotFoundError, ScryfallClient
from mtg_print.sheet import SheetGenerator
from mtg_print.terminal import display_image, display_images_horizontal

app = typer.Typer(no_args_is_help=True)


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"mtg-print {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: Annotated[
        bool, typer.Option("--version", "-v", callback=version_callback, is_eager=True)
    ] = False,
) -> None:
    """MTG proxy printing tool"""


def select_printing_interactive(
    card_name: str,
    printings: list[CardPrinting],
    client: ScryfallClient,
    index: int,
    total: int,
    show_thumbnails: bool = True,
) -> CardPrinting:
    if len(printings) == 1:
        return printings[0]

    typer.echo(f"\n[{index}/{total}] {card_name} ({len(printings)} printings)\n")

    for i, p in enumerate(printings, 1):
        default = " [default]" if i == 1 else ""
        dfc = " (DFC)" if p.is_double_faced else ""
        typer.echo(
            f"  {i}. {p.set_code.upper()} ({p.release_date.year}) - {p.set_name}{dfc}{default}"
        )

    if show_thumbnails:
        typer.echo()
        thumbnails: list[bytes] = []
        for p in printings:
            if p.faces and p.faces[0].image_uri_small:
                response = client.client.get(p.faces[0].image_uri_small)
                if response.status_code == 200:
                    thumbnails.append(response.content)
        if thumbnails:
            display_images_horizontal(thumbnails)

    typer.echo()
    choice = typer.prompt("Select printing", default="1", show_default=False)

    selected: CardPrinting
    try:
        idx = int(choice) - 1
        if 0 <= idx < len(printings):
            selected = printings[idx]
        else:
            typer.echo("Invalid selection, using default")
            selected = printings[0]
    except ValueError:
        typer.echo("Invalid selection, using default")
        selected = printings[0]

    save_preference(card_name, selected.set_code)
    return selected


@app.command()
def build(
    decklist_path: Annotated[Path, typer.Argument(help="Path to decklist file")],
    output: Annotated[Path | None, typer.Option("--output", "-o", help="Output PDF path")] = None,
    set_override: Annotated[
        list[str], typer.Option("--set", "-s", help="Override set for card: 'Card Name=SET'")
    ] = [],
    interactive: Annotated[
        bool, typer.Option("--interactive", "-i", help="Interactively select card art")
    ] = False,
    extras: Annotated[
        bool, typer.Option("--extras", "-e", help="Include tokens, emblems, and related cards")
    ] = False,
) -> None:
    """Build printable PDF from decklist."""
    overrides = {}
    for override in set_override:
        if "=" in override:
            name, set_code = override.rsplit("=", 1)
            overrides[name.strip()] = set_code.strip()

    decklist = parse_decklist(decklist_path)
    typer.echo(f"Parsed {len(decklist.entries)} unique cards ({decklist.total_cards} total)")

    client = ScryfallClient()
    cache = ImageCache()
    images: list[Path] = []
    processed_entries: list[tuple[str, str | None]] = []
    total_cards = len(decklist.entries)

    if interactive:
        for idx, entry in enumerate(decklist.entries, 1):
            set_code = overrides.get(entry.name, entry.set_override)
            pref = get_preference(entry.name)
            try:
                if set_code:
                    printing = client.get_card_by_name(entry.name, set_code)
                elif pref:
                    printing = client.get_card_by_name(entry.name, pref)
                    typer.echo(f"[{idx}/{total_cards}] {entry.name} -> {pref.upper()} (saved)")
                else:
                    printings = client.search_printings(entry.name)
                    if not printings:
                        raise CardNotFoundError(entry.name)
                    printing = select_printing_interactive(
                        entry.name, printings, client, idx, total_cards
                    )
            except CardNotFoundError:
                typer.echo(f"\nCard not found: '{entry.name}'", err=True)
                raise typer.Exit(1)

            for _ in range(entry.count):
                front = cache.get_or_download(printing, client, face_index=0)
                images.append(front)
                if printing.is_double_faced:
                    back = cache.get_or_download(printing, client, face_index=1)
                    images.append(back)

            processed_entries.append((entry.name, set_code))
    else:
        with typer.progressbar(decklist.entries, label="Fetching cards") as entries:
            for entry in entries:
                set_code = (
                    overrides.get(entry.name) or get_preference(entry.name) or entry.set_override
                )
                try:
                    printing = client.get_card_by_name(entry.name, set_code)
                except CardNotFoundError:
                    typer.echo(f"\nCard not found: '{entry.name}'", err=True)
                    raise typer.Exit(1)

                for _ in range(entry.count):
                    front = cache.get_or_download(printing, client, face_index=0)
                    images.append(front)
                    if printing.is_double_faced:
                        back = cache.get_or_download(printing, client, face_index=1)
                        images.append(back)

                processed_entries.append((entry.name, set_code))

    if extras:
        related_parts: dict[str, CardPrinting] = {}
        for name, set_code in processed_entries:
            for part in client.get_related_parts(name, set_code):
                if part.name not in related_parts:
                    related_parts[part.name] = part

        if related_parts:
            typer.echo(f"Found {len(related_parts)} related tokens/emblems")
            for part in related_parts.values():
                front = cache.get_or_download(part, client, face_index=0)
                images.append(front)
                if part.is_double_faced:
                    back = cache.get_or_download(part, client, face_index=1)
                    images.append(back)

    output_path = output or decklist_path.with_suffix(".pdf")
    generator = SheetGenerator()
    generator.generate(images, output_path)
    typer.echo(f"\nGenerated {output_path} with {len(images)} card images")


@app.command()
def search(
    card_name: Annotated[str, typer.Argument(help="Card name to search")],
    limit: Annotated[int, typer.Option("--limit", "-n", help="Max results")] = 10,
    preview: Annotated[bool, typer.Option("--preview", "-p", help="Show card thumbnails")] = False,
) -> None:
    """Search for card printings (oldest first)."""
    client = ScryfallClient()
    try:
        printings = client.search_printings(card_name)
    except CardNotFoundError:
        typer.echo(f"No card found matching '{card_name}'", err=True)
        raise typer.Exit(1)

    for i, p in enumerate(printings[:limit], 1):
        dfc = " (DFC)" if p.is_double_faced else ""
        typer.echo(f"  {i}. {p.set_code.upper()} ({p.release_date}) - {p.set_name}{dfc}")

        if preview and p.faces and p.faces[0].image_uri_small:
            response = client.client.get(p.faces[0].image_uri_small)
            if response.status_code == 200:
                display_image(response.content, width=15)


@app.command()
def cache(
    clear: Annotated[bool, typer.Option("--clear", help="Clear image cache")] = False,
    stats: Annotated[bool, typer.Option("--stats", help="Show cache statistics")] = False,
) -> None:
    """Manage image cache."""
    image_cache = ImageCache()

    if clear:
        count = image_cache.clear()
        typer.echo(f"Cleared {count} cached images")
    elif stats:
        s = image_cache.stats()
        size_mb = s["total_size_bytes"] / (1024 * 1024)
        typer.echo(f"Cached images: {s['file_count']}")
        typer.echo(f"Total size: {size_mb:.2f} MB")
    else:
        typer.echo("Use --clear or --stats")


@app.command()
def prefs(
    clear: Annotated[bool, typer.Option("--clear", help="Clear all saved preferences")] = False,
) -> None:
    """Manage card art preferences."""
    if clear:
        count = clear_preferences()
        typer.echo(f"Cleared {count} saved preferences")
    else:
        preferences = load_preferences()
        if not preferences:
            typer.echo("No saved preferences")
        else:
            typer.echo(f"Saved preferences ({len(preferences)} cards):\n")
            for card_name, set_code in sorted(preferences.items()):
                typer.echo(f"  {card_name} -> {set_code.upper()}")


if __name__ == "__main__":
    app()



import base64
import io
import shutil
import sys

from PIL import Image, ImageDraw, ImageFont


def get_terminal_width() -> int:
    return shutil.get_terminal_size().columns


def display_image_iterm(image_data: bytes, width: int = 20) -> None:
    b64 = base64.b64encode(image_data).decode("ascii")
    sys.stdout.write(f"\033]1337;File=inline=1;width={width}:{b64}\a\n")
    sys.stdout.flush()


def display_image(image_data: bytes, width: int = 20) -> bool:
    try:
        display_image_iterm(image_data, width)
        return True
    except Exception:
        return False


def composite_images_horizontal(
    images: list[bytes], spacing_px: int = 10, max_per_row: int = 6
) -> bytes:
    pil_images = [Image.open(io.BytesIO(data)) for data in images]
    if not pil_images:
        raise ValueError("No images to composite")

    label_height = 20
    img_width = pil_images[0].width
    img_height = pil_images[0].height

    num_rows = (len(pil_images) + max_per_row - 1) // max_per_row
    images_in_widest_row = min(len(pil_images), max_per_row)

    total_width = images_in_widest_row * img_width + spacing_px * (images_in_widest_row - 1)
    total_height = num_rows * (img_height + label_height) + spacing_px * (num_rows - 1)

    composite = Image.new("RGB", (total_width, total_height), color=(30, 30, 30))
    draw = ImageDraw.Draw(composite)

    try:
        font = ImageFont.truetype("/System/Library/Fonts/Helvetica.ttc", 14)
    except OSError:
        font = ImageFont.load_default()

    for i, img in enumerate(pil_images):
        row = i // max_per_row
        col = i % max_per_row

        x = col * (img_width + spacing_px)
        y = row * (img_height + label_height + spacing_px)

        composite.paste(img, (x, y))

        label = str(i + 1)
        bbox = draw.textbbox((0, 0), label, font=font)
        text_width = bbox[2] - bbox[0]
        text_x = x + (img_width - text_width) // 2
        text_y = y + img_height + 2
        draw.text((text_x, text_y), label, fill=(255, 255, 255), font=font)

    output = io.BytesIO()
    composite.save(output, format="JPEG", quality=85)
    return output.getvalue()


def display_images_horizontal(images: list[bytes], spacing_px: int = 10) -> bool:
    if not images:
        return True

    try:
        composite_data = composite_images_horizontal(images, spacing_px)
        terminal_width = get_terminal_width()
        return display_image(composite_data, width=terminal_width)
    except Exception:
        return False



from datetime import date

import pytest

from mtg_print.models import CardFace, CardPrinting, DeckEntry, Decklist


class TestCardPrinting:
    @pytest.mark.parametrize(
        "layout,expected",
        [
            ("transform", True),
            ("modal_dfc", True),
            ("reversible_card", True),
            ("normal", False),
            ("adventure", False),
            ("split", False),
        ],
    )
    def test_is_double_faced(self, layout: str, expected: bool) -> None:
        printing = CardPrinting(
            name="Test",
            set_code="tst",
            set_name="Test Set",
            collector_number="1",
            release_date=date(2020, 1, 1),
            scryfall_uri="https://scryfall.com/card/tst/1",
            layout=layout,
            faces=[],
        )
        assert printing.is_double_faced is expected

    def test_front_image(self) -> None:
        printing = CardPrinting(
            name="Test",
            set_code="tst",
            set_name="Test Set",
            collector_number="1",
            release_date=date(2020, 1, 1),
            scryfall_uri="https://scryfall.com/card/tst/1",
            layout="normal",
            faces=[CardFace(name="Test", image_uri_png="https://example.com/front.png")],
        )
        assert printing.front_image == "https://example.com/front.png"

    def test_back_image_when_present(self) -> None:
        printing = CardPrinting(
            name="Test // Back",
            set_code="tst",
            set_name="Test Set",
            collector_number="1",
            release_date=date(2020, 1, 1),
            scryfall_uri="https://scryfall.com/card/tst/1",
            layout="transform",
            faces=[
                CardFace(name="Test", image_uri_png="https://example.com/front.png"),
                CardFace(name="Back", image_uri_png="https://example.com/back.png"),
            ],
        )
        assert printing.back_image == "https://example.com/back.png"

    def test_back_image_when_absent(self) -> None:
        printing = CardPrinting(
            name="Test",
            set_code="tst",
            set_name="Test Set",
            collector_number="1",
            release_date=date(2020, 1, 1),
            scryfall_uri="https://scryfall.com/card/tst/1",
            layout="normal",
            faces=[CardFace(name="Test", image_uri_png="https://example.com/front.png")],
        )
        assert printing.back_image is None


class TestDecklist:
    def test_total_cards_empty(self) -> None:
        decklist = Decklist(entries=[])
        assert decklist.total_cards == 0

    def test_total_cards_single_entry(self) -> None:
        decklist = Decklist(entries=[DeckEntry(count=4, name="Lightning Bolt")])
        assert decklist.total_cards == 4

    def test_total_cards_multiple_entries(self) -> None:
        decklist = Decklist(
            entries=[
                DeckEntry(count=4, name="Lightning Bolt"),
                DeckEntry(count=4, name="Elvish Reclaimer"),
                DeckEntry(count=2, name="Sylvan Library"),
            ]
        )
        assert decklist.total_cards == 10


class TestDeckEntry:
    def test_set_override_optional(self) -> None:
        entry = DeckEntry(count=4, name="Test Card")
        assert entry.set_override is None

    def test_set_override_provided(self) -> None:
        entry = DeckEntry(count=4, name="Test Card", set_override="MH1")
        assert entry.set_override == "MH1"
